#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.trigger module

This modules defines tasks triggers, matching selected scheduling modes.
"""

from datetime import timedelta

from apscheduler.triggers.base import BaseTrigger
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger
from persistent import Persistent
from zope.componentvocabulary.vocabulary import UtilityVocabulary
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces import TASK_SCHEDULING_MODES_VOCABULARY
from pyams_scheduler.interfaces.task import ICronTask, ICronTaskScheduling, IDateTask, \
    IDateTaskScheduling, ILoopTask, ILoopTaskScheduling, ITaskSchedulingMode, \
    SCHEDULER_TASK_CRON_INFO, SCHEDULER_TASK_DATE_INFO, SCHEDULER_TASK_LOOP_INFO
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.date import date_to_datetime
from pyams_utils.factory import factory_config
from pyams_utils.registry import utility_config
from pyams_utils.timezone import tztime
from pyams_utils.vocabulary import vocabulary_config


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@vocabulary_config(name=TASK_SCHEDULING_MODES_VOCABULARY)
class SchedulingModesVocabulary(UtilityVocabulary):
    """Scheduling modes vocabulary"""

    interface = ITaskSchedulingMode
    nameOnly = True


#
# Immediate task trigger
#

class ImmediateTaskTrigger(BaseTrigger):  # pylint: disable=no-init
    """Immediate-style task scheduler"""

    def get_next_fire_time(self, previous_fire_time, now):
        """Get next task fire time"""
        if previous_fire_time:
            return None
        return now + timedelta(seconds=5)


#
# Cron-style scheduling mode
#

@factory_config(ICronTaskScheduling)
class CronTaskScheduleInfo(Persistent):
    """Cron-style schedule info"""

    active = FieldProperty(ICronTaskScheduling['active'])
    start_date = FieldProperty(ICronTaskScheduling['start_date'])
    end_date = FieldProperty(ICronTaskScheduling['end_date'])
    year = FieldProperty(ICronTaskScheduling['year'])
    month = FieldProperty(ICronTaskScheduling['month'])
    day = FieldProperty(ICronTaskScheduling['day'])
    week = FieldProperty(ICronTaskScheduling['week'])
    day_of_week = FieldProperty(ICronTaskScheduling['day_of_week'])
    hour = FieldProperty(ICronTaskScheduling['hour'])
    minute = FieldProperty(ICronTaskScheduling['minute'])
    second = FieldProperty(ICronTaskScheduling['second'])


@adapter_config(required=ICronTask, provides=ICronTaskScheduling)
def cron_task_scheduler_info_factory(context):
    """Cron-style task scheduling info factory"""
    return get_annotation_adapter(context, SCHEDULER_TASK_CRON_INFO, ICronTaskScheduling,
                                  notify=False, locate=False)


@utility_config(name='Cron-style scheduling', provides=ITaskSchedulingMode)
class CronTaskScheduler:
    """Cron-style scheduler mode"""

    marker_interface = ICronTask
    schema = ICronTaskScheduling

    def get_trigger(self, task):
        """Get cron-like task trigger"""
        if not self.marker_interface.providedBy(task):
            raise Exception(_("Task is not configured for cron-style scheduling!"))
        info = self.schema(task, None)
        if info is None:
            return None
        return CronTrigger(year=info.year or '*',
                           month=info.month or '*',
                           day=info.day or '*',
                           week=info.week or '*',
                           day_of_week=info.day_of_week or '*',
                           hour=info.hour or '*',
                           minute=info.minute or '*',
                           second=info.second or '0',
                           start_date=tztime(date_to_datetime(info.start_date)),
                           end_date=tztime(date_to_datetime(info.end_date)))


#
# Date-style scheduling mode
#

@factory_config(IDateTaskScheduling)
class DateTaskScheduleInfo(Persistent):
    """Date-style schedule info"""

    active = FieldProperty(IDateTaskScheduling['active'])
    start_date = FieldProperty(IDateTaskScheduling['start_date'])


@adapter_config(required=IDateTask, provides=IDateTaskScheduling)
def date_task_scheduler_info_factory(context):
    """Date-style task scheduling info factory"""
    return get_annotation_adapter(context, SCHEDULER_TASK_DATE_INFO, IDateTaskScheduling,
                                  notify=False, locate=False)


@utility_config(name='Date-style scheduling', provides=ITaskSchedulingMode)
class DateTaskScheduler:
    """Date-style scheduler mode"""

    marker_interface = IDateTask
    schema = IDateTaskScheduling

    def get_trigger(self, task):
        """Get date trigger"""
        if not self.marker_interface.providedBy(task):
            raise Exception(_("Task is not configured for date-style scheduling!"))
        info = self.schema(task, None)
        if info is None:
            return None
        return DateTrigger(run_date=tztime(date_to_datetime(info.start_date)))


#
# Loop-style scheduling mode
#

@factory_config(ILoopTaskScheduling)
class LoopTaskScheduleInfo(Persistent):
    """Loop-style schedule info"""

    active = FieldProperty(ILoopTaskScheduling['active'])
    start_date = FieldProperty(ILoopTaskScheduling['start_date'])
    end_date = FieldProperty(ILoopTaskScheduling['end_date'])
    weeks = FieldProperty(ILoopTaskScheduling['weeks'])
    days = FieldProperty(ILoopTaskScheduling['days'])
    hours = FieldProperty(ILoopTaskScheduling['hours'])
    minutes = FieldProperty(ILoopTaskScheduling['minutes'])
    seconds = FieldProperty(ILoopTaskScheduling['seconds'])


@adapter_config(required=ILoopTask, provides=ILoopTaskScheduling)
def loop_task_scheduler_info_factory(context):
    """Loop-style task scheduling info factory"""
    return get_annotation_adapter(context, SCHEDULER_TASK_LOOP_INFO, ILoopTaskScheduling,
                                  notify=False, locate=False)


@utility_config(name='Loop-style scheduling', provides=ITaskSchedulingMode)
class LoopTaskScheduler:
    """Loop-style scheduler mode"""

    marker_interface = ILoopTask
    schema = ILoopTaskScheduling

    def get_trigger(self, task):
        """Get loop interval trigger"""
        if not self.marker_interface.providedBy(task):
            raise Exception(_("Task is not configured for loop-style scheduling!"))
        info = self.schema(task, None)
        if info is None:
            return None
        return IntervalTrigger(weeks=info.weeks,
                               days=info.days,
                               hours=info.hours,
                               minutes=info.minutes,
                               seconds=info.seconds,
                               start_date=tztime(date_to_datetime(info.start_date)),
                               end_date=tztime(date_to_datetime(info.end_date)))



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.task.sync module

This module defines the management interface components of directory
synchronization tasks.
"""

from zope.component import getAdapters
from zope.interface import Interface, implementer, implementer_only

from pyams_form.ajax import ajax_form_config
from pyams_form.browser.widget import HTMLFormElement
from pyams_form.converter import BaseDataConverter
from pyams_form.field import Fields
from pyams_form.group import GroupManager
from pyams_form.interfaces import IDataConverter, IObjectFactory
from pyams_form.interfaces.form import IForm, IInnerTabForm
from pyams_form.interfaces.widget import IFieldWidget, IObjectWidget
from pyams_form.subform import InnerAddForm, InnerEditForm
from pyams_form.widget import FieldWidget, Widget
from pyams_layer.interfaces import IFormLayer, IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task.sync import IDirectoryHandler, IDirectoryHandlerHostField, \
    IDirectoryInfo, IDirectorySyncTask, IDirectorySyncTaskInfo
from pyams_scheduler.task.sync import DirectoryInfo
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_scheduler.zmi.interfaces import IDirectoryHandlerHostWidget
from pyams_scheduler.zmi.task import BaseTaskAddForm, BaseTaskEditForm
from pyams_skin.viewlet.menu import MenuItem
from pyams_utils.adapter import adapter_config
from pyams_utils.factory import get_interface_name
from pyams_utils.interfaces.form import NO_VALUE
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@adapter_config(name=get_interface_name(IDirectoryInfo),
                required=(Interface, IFormLayer, IForm, IObjectWidget),
                provides=IObjectFactory)
def directory_info_factory(*args):  # pylint: disable=unused-argument
    """Directory info object factory"""
    return DirectoryInfo


#
# Directory handler host widget
#

@adapter_config(required=(IDirectoryHandlerHostField, IDirectoryHandlerHostWidget),
                provides=IDataConverter)
class DirectoryHandlerHostDataConverter(BaseDataConverter):
    """Directory handler host data converter"""

    def to_widget_value(self, value):
        return value

    def to_field_value(self, value):
        return value


@implementer_only(IDirectoryHandlerHostWidget)
class DirectoryHandlerHostWidget(HTMLFormElement, Widget):
    """Directory handler host getter widget"""

    @property
    def display_value(self):
        """Display value getter"""
        return self.value or (None, None)

    def extract(self, default=NO_VALUE):
        params = self.request.params
        marker = params.get('{}-empty-marker'.format(self.name), default)
        if marker is not default:
            protocol = params.get('{}-protocol'.format(self.name))
            hostname = params.get('{}-hostname'.format(self.name))
            return (protocol, hostname) if protocol else default
        return default

    @property
    def protocols(self):
        """Getter of supported protocols"""
        info = DirectoryInfo()
        yield from sorted(name for name, adapter in getAdapters((info,), IDirectoryHandler))


@adapter_config(required=(IDirectoryHandlerHostField, IFormLayer),
                provides=IFieldWidget)
def DirectoryHandlerHostFieldWidget(field, request):  # pylint: disable=invalid-name
    """Directory handler host widget factory"""
    return FieldWidget(field, DirectoryHandlerHostWidget(request))


#
# Directory synchronization base form
#

class IDirectorySyncTaskForm(IForm):
    """Directory synchronization forms interface"""


@implementer(IDirectorySyncTaskForm)
class DirectorySyncTaskFormInfo(GroupManager):
    """Directory synchronization task form info"""

    title = _("Synchronization settings")

    fields = Fields(IDirectorySyncTaskInfo)


#
# Directory synchronization task add form
#

@viewlet_config(name='add-sync-task.menu',
                context=IScheduler, layer=IAdminLayer, view=SchedulerTasksTable,
                manager=IContextAddingsViewletManager, weight=40,
                permission=MANAGE_TASKS_PERMISSION)
class DirectorySyncTaskAddMenu(MenuItem):
    """Directory synchronization task add menu"""

    label = _("Add directory synchronization...")
    href = 'add-sync-task.html'
    modal_target = True


@ajax_form_config(name='add-sync-task.html', context=IScheduler, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class DirectorySyncTaskAddForm(BaseTaskAddForm):
    """Directory synchronization task add form"""

    content_factory = IDirectorySyncTask


@adapter_config(name='sync-task-info.form',
                required=(IScheduler, IAdminLayer, DirectorySyncTaskAddForm),
                provides=IInnerTabForm)
class DirectorySyncTaskAddFormInfo(DirectorySyncTaskFormInfo, InnerAddForm):
    """Directory synchronization task add form info"""


#
# Directory synchronization task edit form
#

@ajax_form_config(name='properties.html', context=IDirectorySyncTask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class DirectorySyncTaskEditForm(BaseTaskEditForm):
    """Directory synchronization task edit form"""


@adapter_config(name='sync-task-info.form',
                required=(IDirectorySyncTask, IAdminLayer, DirectorySyncTaskEditForm),
                provides=IInnerTabForm)
class DirectorySyncTaskEditFormInfo(DirectorySyncTaskFormInfo, InnerEditForm):
    """Directory synchronization task edit form info"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task module

This module defines base tasks classes.
"""

__docformat__ = 'restructuredtext'

import logging
import traceback
from datetime import datetime
from io import StringIO

import transaction
from persistent import Persistent
from pyramid.config import Configurator
from pyramid.events import subscriber
from pyramid.threadlocal import RequestContext, manager
from transaction.interfaces import ITransactionManager
from zope.component import queryUtility
from zope.component.interfaces import ISite
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.copy.interfaces import ICopyHook, ResumeCopy
from zope.interface import alsoProvides, implementer, noLongerProvides
from zope.intid import IIntIds
from zope.lifecycleevent import IObjectAddedEvent, IObjectModifiedEvent, IObjectRemovedEvent
from zope.location import locate
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces import AfterRunJobEvent, BeforeRunJobEvent, IScheduler, ITask, \
    ITaskHistory, MANAGE_TASKS_PERMISSION, SCHEDULER_AUTH_KEY, SCHEDULER_HANDLER_KEY, \
    SCHEDULER_NAME
from pyams_scheduler.interfaces.task import ITaskHistoryContainer, ITaskInfo, \
    ITaskNotificationContainer, ITaskSchedulingMode, TASK_STATUS_EMPTY, \
    TASK_STATUS_ERROR, TASK_STATUS_NONE, TASK_STATUS_OK
from pyams_security.interfaces import IViewContextPermissionChecker
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.date import get_duration
from pyams_utils.registry import get_pyramid_registry, get_utility, query_utility, \
    set_local_registry
from pyams_utils.request import check_request
from pyams_utils.timezone import tztime
from pyams_utils.traversing import get_parent
from pyams_utils.zodb import ZODBConnection
from pyams_zmq.socket import zmq_response, zmq_socket


LOGGER = logging.getLogger('PyAMS (scheduler)')


@implementer(ITaskHistory)
class TaskHistoryItem(Persistent, Contained):
    """Task history item"""

    status = FieldProperty(ITaskHistory['status'])
    date = FieldProperty(ITaskHistory['date'])
    duration = FieldProperty(ITaskHistory['duration'])
    report = FieldProperty(ITaskHistory['report'])

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


@implementer(ITaskHistoryContainer)
class TaskHistoryContainer(Folder):
    """Task history container"""

    def check_history(self, duration, length):
        """Check history container contents"""
        if duration:
            now = tztime(datetime.utcnow())
            for key in list(self.keys()):
                if (now - tztime(self[key].date)).days > duration:
                    del self[key]
        if length and (len(self) > length):
            keys = sorted(self.keys(), reverse=True)[:length]
            for key in list(self.keys()):
                if key not in keys:
                    del self[key]


@implementer(ITask)
class Task(Persistent, Contained):
    """Task definition persistent class"""

    label = None
    icon_class = None

    name = FieldProperty(ITask['name'])
    _schedule_mode = FieldProperty(ITask['schedule_mode'])
    keep_empty_reports = FieldProperty(ITask['keep_empty_reports'])
    _history_duration = FieldProperty(ITask['history_duration'])
    _history_length = FieldProperty(ITask['history_length'])

    settings_view_name = FieldProperty(ITask['settings_view_name'])
    principal_id = None

    _internal_id = None

    def __init__(self):
        history = self.history = TaskHistoryContainer()
        locate(history, self, '++history++')

    @property
    def schedule_mode(self):
        """Scheduler mode getter"""
        return self._schedule_mode

    @schedule_mode.setter
    def schedule_mode(self, value):
        """Scheduler mode setter"""
        if self._schedule_mode is not None:
            mode = query_utility(ITaskSchedulingMode, name=self._schedule_mode)
            if (mode is not None) and mode.marker_interface.providedBy(self):
                noLongerProvides(self, mode.marker_interface)
        self._schedule_mode = value
        if value:
            mode = get_utility(ITaskSchedulingMode, name=value)
            alsoProvides(self, mode.marker_interface)
            mode.schema(self).active = False
            if self.__parent__ is not None:
                self.reset()

    @property
    def history_duration(self):
        """History duration getter"""
        return self._history_duration

    @history_duration.setter
    def history_duration(self, value):
        """History duration setter"""
        self._history_duration = value

    @property
    def history_length(self):
        """History length getter"""
        return self._history_length

    @history_length.setter
    def history_length(self, value):
        """History length setter"""
        self._history_length = value

    def check_history(self):
        """Check history container for old contents"""
        self.history.check_history(self.history_duration, self.history_length)

    @property
    def internal_id(self):
        """Task internal ID getter"""
        if self._internal_id is None:
            site = get_parent(self, ISite)
            sm = site.getSiteManager()  # pylint: disable=invalid-name,too-many-function-args,assignment-from-no-return
            intids = sm.queryUtility(IIntIds)
            if intids is not None:
                self._internal_id = intids.register(self)
        return self._internal_id

    def get_trigger(self):
        """Task trigger getter"""
        mode = queryUtility(ITaskSchedulingMode, self.schedule_mode)
        if mode is None:
            return None
        return mode.get_trigger(self)

    def get_scheduling_info(self):
        """Task scheduling info getter"""
        mode = queryUtility(ITaskSchedulingMode, self.schedule_mode)
        if mode is None:
            return None
        return mode.schema(self, None)

    def reset(self):
        """Task reset launcher"""
        scheduler_util = query_utility(IScheduler)
        if scheduler_util is not None:
            # get task internal ID before transaction ends!!!
            transaction.get().addAfterCommitHook(self._reset_action, kws={
                'scheduler': scheduler_util,
                'registry': get_pyramid_registry(),
                'job_id': self.internal_id
            })

    def _reset_action(self, status, *args, **kwargs):  # pylint: disable=unused-argument
        """Task reset action, called in an after-commit hook"""
        if not status:
            return
        scheduler_util = kwargs.get('scheduler')
        if scheduler_util is None:
            return
        registry = kwargs.get('registry') or get_pyramid_registry()
        handler = registry.settings.get(SCHEDULER_HANDLER_KEY, False)
        if handler:
            zmq_settings = {
                'zodb_name': scheduler_util.zodb_name,
                'task_name': self.__name__,
                'job_id': kwargs.get('job_id')
            }
            LOGGER.debug("Resetting task {0.name} with {1!r}".format(self, zmq_settings))
            socket = zmq_socket(handler, auth=registry.settings.get(SCHEDULER_AUTH_KEY))
            socket.send_json(['reset_task', zmq_settings])
            zmq_response(socket)

    def launch(self):
        """Task immediate launcher"""
        scheduler_util = query_utility(IScheduler)
        if scheduler_util is not None:
            # get task internal ID before transaction ends!!!
            transaction.get().addAfterCommitHook(self._launch_action, kws={
                'scheduler': scheduler_util,
                'registry': get_pyramid_registry(),
                'job_id': self.internal_id
            })

    def _launch_action(self, status, *args, **kwargs):  # pylint: disable=unused-argument
        """Task immediate launch action"""
        if not status:
            return
        scheduler_util = kwargs.get('scheduler')
        if scheduler_util is None:
            return
        registry = kwargs.get('registry') or get_pyramid_registry()
        handler = registry.settings.get(SCHEDULER_HANDLER_KEY, False)
        if handler:
            zmq_settings = {
                'zodb_name': scheduler_util.zodb_name,
                'task_name': self.__name__,
                'job_id': kwargs.get('job_id')
            }
            LOGGER.debug("Running task {0.name} with {1!r}".format(self, zmq_settings))
            socket = zmq_socket(handler, auth=registry.settings.get(SCHEDULER_AUTH_KEY))
            socket.send_json(['run_task', zmq_settings])
            zmq_response(socket)

    def __call__(self, *args, **kwargs):
        report = StringIO()
        return self._run(report, **kwargs)

    def is_runnable(self):
        """Task runnable state getter"""
        mode = queryUtility(ITaskSchedulingMode, self.schedule_mode)
        if mode is None:
            return False
        info = mode.schema(self, None)
        if info is None:
            return False
        return info.active

    def _run(self, report, **kwargs):  # pylint: disable=too-many-locals
        """Task execution wrapper"""
        status = TASK_STATUS_NONE
        result = None
        # initialize ZCA hook
        registry = kwargs.pop('registry')
        manager.push({'registry': registry, 'request': None})
        config = Configurator(registry=registry)
        config.hook_zca()
        # open ZODB connection
        zodb_connection = ZODBConnection(name=kwargs.pop('zodb_name', ''))
        with zodb_connection as root:
            try:
                application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                         PYAMS_APPLICATION_DEFAULT_NAME)
                sm = root.get(application_name).getSiteManager()  # pylint: disable=invalid-name
                scheduler_util = sm.get(SCHEDULER_NAME)
                task = scheduler_util.get(self.__name__)
                if task is not None:
                    set_local_registry(sm)
                    request = check_request(registry=registry, principal_id=self.principal_id)
                    with RequestContext(request):
                        if not (kwargs.get('run_immediate') or task.is_runnable()):
                            LOGGER.debug("Skipping inactive task {0}".format(task.name))
                            return status, result
                        tm = ITransactionManager(task)  # pylint: disable=invalid-name
                        for attempt in tm.attempts():
                            with attempt as t:  # pylint: disable=invalid-name
                                start_date = datetime.utcnow()
                                duration = 0.
                                try:
                                    registry.notify(BeforeRunJobEvent(task))
                                    (status, result) = task.run(report, **kwargs)
                                    end_date = datetime.utcnow()
                                    duration = (end_date - start_date).total_seconds()
                                    report.write('\n\nTask duration: {0}'.format(
                                        get_duration(start_date, request=request)))
                                except:  # pylint: disable=bare-except
                                    status = TASK_STATUS_ERROR
                                    # pylint: disable=protected-access
                                    task._log_exception(report,
                                                        "An error occurred during execution of "
                                                        "task '{0}'".format(task.name))
                                registry.notify(AfterRunJobEvent(task, status, result))
                                task.store_report(report, status, start_date, duration)
                                task.send_report(report, status, registry)
                            if t.status == 'Committed':
                                break
            except:  # pylint: disable=bare-except
                self._log_exception(None, "Can't execute scheduled job {0}".format(self.name))
            tm = ITransactionManager(self, None)  # pylint: disable=invalid-name
            if tm is not None:
                tm.abort()
        return status, result

    def run(self, report, **kwargs):  # pylint: disable=no-self-use
        """Task run implementation

        May result a tuple containing a status code and a result (which can be empty).
        """
        raise NotImplementedError("The 'run' method must be implemented by Task subclasses!")

    @staticmethod
    def _log_report(report, message, add_timestamp=True, level=logging.INFO):
        """Execution log report"""
        if isinstance(message, bytes):
            message = message.decode()
        if add_timestamp:
            message = '{0} - {1}'.format(tztime(datetime.utcnow()).strftime('%c'), message)
        if report is not None:
            report.write(message + '\n')
        LOGGER.log(level, message)

    @staticmethod
    def _log_exception(report, message=None):
        """Exception log report"""
        if isinstance(message, bytes):
            message = message.decode()
        message = '{0} - {1}'.format(tztime(datetime.utcnow()).strftime('%c'),
                                     message or 'An error occurred') + '\n\n'
        if report is not None:
            report.write(message)
            report.write(traceback.format_exc() + '\n')
        LOGGER.exception(message)

    def store_report(self, report, status, start_date, duration):
        """Execution report store"""
        if (status in (TASK_STATUS_NONE, TASK_STATUS_EMPTY)) and \
                not self.keep_empty_reports:
            return
        item = TaskHistoryItem(status=str(status),
                               date=start_date,
                               duration=duration,
                               report=report.getvalue())
        self.history[item.date.isoformat()] = item
        self.check_history()

    def send_report(self, report, status, registry):
        """Execution report messaging"""
        notifications = ITaskNotificationContainer(self)
        for target in notifications.get_enabled_items():
            handler = target.get_handler()
            if handler is None:
                continue
            if ((status in (TASK_STATUS_NONE, TASK_STATUS_EMPTY)) and
                not target.send_empty_reports) or \
                    ((status == TASK_STATUS_OK) and target.report_errors_only):
                return

            handler.send_report(self, report, status, target, registry)


@subscriber(IObjectAddedEvent, context_selector=ITask)
def handle_new_task(event):
    """Handle new task"""
    event.object.reset()


@subscriber(IObjectModifiedEvent, context_selector=ITask)
def handle_modified_task(event):
    """Handle modified task"""
    for changes in event.descriptions:
        if (changes.interface == ITaskInfo) and \
                (('history_duration' in changes.attributes) or
                 ('history_length' in changes.attributes)):
            event.object.check_history()
            break


@subscriber(IObjectRemovedEvent, context_selector=ITask)
def handle_removed_task(event):
    """Handle removed task"""
    request = check_request()
    if request.registry:
        handler = request.registry.settings.get(SCHEDULER_HANDLER_KEY, False)
        if handler:
            task = event.object
            scheduler_util = query_utility(IScheduler)
            zmq_settings = {
                'zodb_name': scheduler_util.zodb_name,
                'task_name': task.__name__,
                'job_id': task.internal_id
            }
            LOGGER.debug("Removing task {0.name} with {1!r}".format(task, zmq_settings))
            socket = zmq_socket(handler, auth=request.registry.settings.get(SCHEDULER_AUTH_KEY))
            socket.send_json(['remove_task', zmq_settings])
            zmq_response(socket)


@adapter_config(required=ITask,
                provides=ICopyHook)
class TaskCopyHook(ContextAdapter):
    """Task copy hook"""

    def __call__(self, toplevel, register):
        register(self._copy_history)
        raise ResumeCopy

    def _copy_history(self, translate):
        task = translate(self.context)
        # create empty history
        history = task.history = TaskHistoryContainer()
        locate(history, task, '++history++')
        # disable task
        scheduling_mode = task.get_scheduling_info()
        scheduling_mode.active = False


@adapter_config(required=ITask,
                provides=IViewContextPermissionChecker)
class TaskPermissionChecker(ContextAdapter):
    """Task permission checker"""

    edit_permission = MANAGE_TASKS_PERMISSION



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.notification module

This module defines mail notification for scheduler tasks.
"""

from persistent import Persistent
from pyramid_mailer import IMailer
from pyramid_mailer.message import Message
from zope.component import queryUtility
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.interface import implementer
from zope.interface.interfaces import ComponentLookupError
from zope.schema.fieldproperty import FieldProperty
from zope.traversing.interfaces import ITraversable

from pyams_scheduler.interfaces import IScheduler, ITask, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task import IMailNotification, ITaskNotification, \
    ITaskNotificationContainer, ITaskNotificationMode, SCHEDULER_TASK_NOTIFICATIONS_KEY, \
    TASK_STATUS_ERROR, TASK_STATUS_WARNING
from pyams_security.interfaces import IViewContextPermissionChecker
from pyams_utils.adapter import ContextAdapter, adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


#
# Task notifications container
#
from pyams_utils.traversing import get_parent


@factory_config(ITaskNotificationContainer)
class TaskNotificationContainer(Folder):
    """Task notification container"""

    def get_enabled_items(self):
        """Get iterator over enabled notifications"""
        yield from (notification for notification in self.values() if notification.enabled)


@adapter_config(required=ITask,
                provides=ITaskNotificationContainer)
def task_notifications_container_factory(context):
    """Task notifications container factory"""
    return get_annotation_adapter(context, SCHEDULER_TASK_NOTIFICATIONS_KEY,
                                  ITaskNotificationContainer,
                                  name='++notify++')


@adapter_config(name='notify',
                required=ITask,
                provides=ITraversable)
class TaskNotificationsTraverser(ContextAdapter):
    """Task notifications traverser"""

    def traverse(self, name, furtherpath=None):
        """Traverser to task notifications"""
        container = ITaskNotificationContainer(self.context)
        if name:
            return container.get(name)
        return container


#
# Base task notifications classes
#

@implementer(ITaskNotification)
class TaskNotification(Persistent, Contained):
    """Base task notification info"""

    enabled = FieldProperty(ITaskNotification['enabled'])
    report_errors_only = FieldProperty(ITaskNotification['report_errors_only'])
    send_empty_reports = FieldProperty(ITaskNotification['send_empty_reports'])

    def get_handler(self):
        """Get notification handler utility"""
        return queryUtility(ITaskNotificationMode, name=self.mode)


@adapter_config(required=ITaskNotification,
                provides=IViewContextPermissionChecker)
class TaskNotificationPermissionChecker(ContextAdapter):
    """Task notification permission checker"""

    edit_permission = MANAGE_TASKS_PERMISSION


#
# Mail notifications classes
#

MAIL_NOTIFICATION_MODE = 'mail'


@factory_config(IMailNotification)
class MailNotification(TaskNotification):
    """Mail notification info"""

    mode = MAIL_NOTIFICATION_MODE

    target_email = FieldProperty(IMailNotification['target_email'])


@utility_config(name=MAIL_NOTIFICATION_MODE,
                provides=ITaskNotificationMode)
class MailNotificationMode:
    """Scheduler task mail notification mode"""

    @staticmethod
    def send_report(task, report, status, target, registry=None):
        """Send mail report to given target"""
        if not IMailNotification.providedBy(target):
            return
        scheduler = get_parent(task, IScheduler)
        try:
            mailer_name = scheduler.report_mailer
        except (TypeError, AttributeError, ComponentLookupError):
            return
        mailer = queryUtility(IMailer, mailer_name)
        if mailer is not None:
            report_source = scheduler.report_source
            if status == TASK_STATUS_ERROR:
                subject = "[SCHEDULER !ERROR!] {}".format(task.name)
            elif status == TASK_STATUS_WARNING:
                subject = "[SCHEDULER WARNING] {}".format(task.name)
            else:
                subject = "[scheduler] {}".format(task.name)
            for target in target.target_email or ():
                message = Message(subject=subject,
                                  sender=report_source,
                                  recipients=(target,),
                                  body=report.getvalue())
                mailer.send(message)
